// PROLL.C: PianoRoll Window

// for Win98 or later.
#define _WIN32_WINDOWS 0x0410

#include <windows.h>
#include "..\res\resource.h"
#include "runtime.h"
#include "smaf.h"
#include "emusmw5.h"	// for main.h:VOICE... TODO:FIX.
#include "main.h"
#include "proll.h"

#define POSBARHEIGHT 12

#define MMT_NONE	0
#define MMT_POSBAR	1
#define MMT_ONROLL	2

// ϕׂȂBmɍŌ゠ĂƂׂ͕Ⴂ񂾂ǁcB
//#define PROLL_ENABLESKIPFIRST

typedef struct{
	UINT RollHeight;
	UINT RollWidth;
	BOOL NiagaraTouch;

	int VisibleTrack;
	UINT CurrentPos;
	SMAF* smaf;
	LPCTSTR OptionString;

	int SoundingNote;	// preview
	EVENT* HilightedEvent;
	EVENT* MovingEvent;
	int MovingPosDelta;
	EVENT* SizingEvent;

	UINT MouseMovingType;

#ifdef PROLL_ENABLESKIPFIRST
	UINT PrePosition;
	EVENT* PreFirstEvent;
#endif

	BOOL ReadOnly;

	// for Debugging?
	BOOL ShowVisibleEventNum;
}PROLLOPTIONSTRUCT;
#define PROLLOPTION (*(PROLLOPTIONSTRUCT*)GetWindowLong(hWnd, 0))

extern MAINOPTION MainOption;

static void draw(HWND hWnd, HDC hdc, int width, int height)
{
	HFONT hfold;
	UINT scrx, scry;
	SMAF* smaf = PROLLOPTION.smaf;	// faster faster...
	PROLLOPTIONSTRUCT* PROption = (PROLLOPTIONSTRUCT*)GetWindowLong(hWnd, 0);
	UINT ProcessEventNum = 0;
	UINT ProcessNoteEventNum = 0;
	UINT VisibleEventNum = 0;

	SetTextColor(hdc, RGB(0, 0, 0));
	SetBkMode(hdc, TRANSPARENT);

	FillRectangle(hdc, 0, 0, width, height, RGB(255, 255, 255));

	{
		LOGFONT lf = {0};

		lf.lfCharSet = DEFAULT_CHARSET;
		strcopy(lf.lfFaceName, "lr SVbN");
		lf.lfHeight = -12;

		hfold = SelectObject(hdc, CreateFontIndirect(&lf));
	}

	{
		SCROLLINFO vsi, hsi;

		vsi.cbSize = hsi.cbSize = sizeof(SCROLLINFO);
		vsi.fMask = hsi.fMask = SIF_POS;
		GetScrollInfo(hWnd, SB_VERT, &vsi);
		GetScrollInfo(hWnd, SB_HORZ, &hsi);

		scrx = hsi.nPos;
		scry = vsi.nPos;
	}

	// draw pianoroll
	{
		int i;
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(220, 220, 220)));
		COLORREF oldcolor = SetTextColor(hdc, RGB(192, 192, 192));
		EVENT* e;

		// draw horizontal grid
		for(i = -1; i < 128; i++)
		{
			MoveToEx(hdc, 0, (127 - i) * PROption->RollHeight - scry + POSBARHEIGHT, NULL);
			LineTo(hdc, width, (127 - i) * PROption->RollHeight - scry + POSBARHEIGHT);

			if(!(i % 12))
			{
				FillRectangle(hdc,	0, (127 - i) * PROption->RollHeight - scry + 1 + POSBARHEIGHT,
									width, (127 - i + 1) * PROption->RollHeight - scry + POSBARHEIGHT, RGB(240, 240, 240));
			}
		}

		// draw vertical grid
		for(i = -((int)scrx * (int)PROption->RollWidth % 100); i < width * (int)PROption->RollWidth; i += 100)
		{
			HPEN hpold;

			if(!((i + scrx * PROption->RollWidth) % (50 * PROption->RollWidth)))	// darker
				hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(192, 192, 192)));

			MoveToEx(hdc, i / PROption->RollWidth, 0 + POSBARHEIGHT, NULL);
			LineTo(hdc, i / PROption->RollWidth, height);

			if(!((i + scrx * PROption->RollWidth) % (50 * PROption->RollWidth)))
				DeleteObject(SelectObject(hdc, hpold));
		}

		if(smaf)
		{
#ifdef PROLL_ENABLESKIPFIRST
			if(scrx * PROption->RollWidth < PROption->PrePosition)
				PROption->PreFirstEvent = NULL;
#endif

			e = smaf->events;

#ifdef PROLL_ENABLESKIPFIRST
			if(PROption->PreFirstEvent)
				e = PROption->PreFirstEvent;

			PROption->PreFirstEvent = NULL;
#endif

			for(; e; e = e->next)
			{
				DWORD gtime;

				ProcessEventNum++;

				if((e->status & 0xF0) != 0x80 && (e->status & 0xF0) != 0x90)	// Not NoteON
					continue;

				ProcessNoteEventNum++;

				if((UINT)((scrx + width) * PROption->RollWidth) < e->time)	// Over. no draw needed any more.
					break;

				if(PROption->VisibleTrack != -1 && (e->status & 0x0F) != PROption->VisibleTrack)	// NoVisible
					continue;

				gtime = getGateTime(e);

				if(e->time + gtime < scrx * PROption->RollWidth)
					continue;

#ifdef PROLL_ENABLESKIPFIRST
				if(!PROption->PreFirstEvent)
				{
					PROption->PreFirstEvent = e;
					PROption->PrePosition = scrx * PROption->RollWidth;
				}
#endif

				VisibleEventNum++;

#if defined(_DEBUG) && 0
				// draw duration line
				Line(hdc,	(e->time - e->duration) / PROption->RollWidth - scrx,
							((127 - e->data[0]) * 2 + 1) * PROption->RollHeight / 2 - scry + POSBARHEIGHT,
							e->time / PROption->RollWidth - scrx,
							((127 - e->data[0]) * 2 + 1) * PROption->RollHeight / 2 - scry + POSBARHEIGHT,
							RGB(128, 0, 128));
#endif

				if(PROption->HilightedEvent == e)
				{
					// Draw border (moving|sizing)
					UINT x = (PROption->HilightedEvent->time + (PROption->HilightedEvent->time + getGateTime(PROption->HilightedEvent)) * 4) / 5;

					Line(hdc,
						x / PROption->RollWidth - scrx, (127 - e->data[0]) * PROption->RollHeight - scry + POSBARHEIGHT - 2,
						x / PROption->RollWidth - scrx, ((127 - e->data[0]) + 1) * PROption->RollHeight - scry + POSBARHEIGHT + 2,
						RGB(255, 0, 0));
				}

				FillRectangle(hdc,	e->time / PROption->RollWidth - scrx,
									(127 - e->data[0]) * PROption->RollHeight - scry + POSBARHEIGHT,
									(e->time + gtime) / PROption->RollWidth - scrx,
									((127 - e->data[0]) + 1) * PROption->RollHeight - scry + POSBARHEIGHT,
									(PROption->HilightedEvent == e) ? RGB(255, 0, 0) : RGB(128, 64, 128));

				FillRectangle(hdc,	e->time / PROption->RollWidth - scrx + 1,
									(127 - e->data[0]) * PROption->RollHeight - scry + 1 + POSBARHEIGHT,
									(e->time + gtime) / PROption->RollWidth - scrx - 1,
									((127 - e->data[0]) + 1) * PROption->RollHeight - scry - 1 + POSBARHEIGHT,
									MainOption.TrackColors[e->status & 0x0F]);
			}

			// Draw info
			if(PROption->HilightedEvent)
			{
				TCHAR ts[80];
				SIZE s;
				LPCTSTR notename[12] = {"C-", "C#", "D-", "D#", "E-", "F-", "F#", "G-", "G#", "A-", "A#", "B-"};
				EVENT* e = PROption->HilightedEvent;
				DWORD gtime = getGateTime(e);
				INT x;

				SaveDC(hdc);

				SetBkMode(hdc, OPAQUE);
				SetBkColor(hdc, GetSysColor(COLOR_WINDOW));
				SetTextColor(hdc, GetSysColor(COLOR_WINDOWTEXT));
				wsprintf(ts, "Track:%u Time:%u Gate:%u %s%d", (e->status & 0x0F) + 1, e->time, gtime, notename[e->data[0] % 12], e->data[0] / 12 - 2);

				GetTextExtentPoint32(hdc, ts, lstrlen(ts), &s);
				x = (e->time/* + gtime*/) / PROption->RollWidth - scrx;
				if(x < 0)
					x = 0;
				TextOut(hdc, x, (127 - e->data[0]) * PROption->RollHeight - scry + POSBARHEIGHT - s.cy, ts, lstrlen(ts));

				RestoreDC(hdc, -1);
			}

			if(PROption->MovingEvent)
			{
				EVENT* e = PROption->MovingEvent;
				DWORD gtime = getGateTime(e);

				Line(hdc,
					e->time / PROption->RollWidth - scrx, 0,
					e->time / PROption->RollWidth - scrx, height,
					RGB(128, 192, 128));
				Line(hdc,
					(e->time + gtime) / PROption->RollWidth - scrx, 0,
					(e->time + gtime) / PROption->RollWidth - scrx, height,
					RGB(128, 192, 128));
			}
			if(PROption->SizingEvent)
			{
				EVENT* e = PROption->SizingEvent;
				DWORD gtime = getGateTime(e);

				Line(hdc,
					e->time / PROption->RollWidth - scrx, 0,
					e->time / PROption->RollWidth - scrx, height,
					RGB(128, 128, 192));
				Line(hdc,
					(e->time + gtime) / PROption->RollWidth - scrx, 0,
					(e->time + gtime) / PROption->RollWidth - scrx, height,
					RGB(128, 128, 192));
			}
		}

		// draw horizontal grid text(ScoreLine)
		for(i = 0; i < 128; i++)
		{
			LPCTSTR oto[] = {TEXT("C"), TEXT("C#"), TEXT("D"), TEXT("D#"), TEXT("E"), TEXT("F"), TEXT("F#"), TEXT("G"), TEXT("G#"), TEXT("A"), TEXT("A#"), TEXT("B")};
			MoveToEx(hdc, 1, (127 - i) * PROption->RollHeight - scry + 1 + POSBARHEIGHT - 2, NULL);
			PutText(hdc, "%-2s %d=%d", oto[i % 12], i / 12 - 2, i);
		}

		SetTextColor(hdc, oldcolor);
		DeleteObject(SelectObject(hdc, hpold));
	}

	// draw PosBar
	{
		COLORREF oldcolor = SetTextColor(hdc, GetSysColor(COLOR_BTNTEXT));
		int i;
		UINT rw = PROption->RollWidth;

		FillRectangle(hdc, 0, 0, width, POSBARHEIGHT, GetSysColor(COLOR_BTNFACE));

		// draw vertical grid text(TimeLine)
		for(i = -((int)scrx * (int)rw % (/*Ȃł50Ȃ100ȂƂ܂Ȃ?->*/100 * (int)rw)); i < width * (int)rw; i += 100)
		{
			if((i + scrx * rw) % (50 * rw))
				continue;
			MoveToEx(hdc, i / (int)PROption->RollWidth + 2, 0, NULL);
			PutText(hdc, "%ums", i + scrx * rw);
		}

		SetTextColor(hdc, oldcolor);
	}

	// Start
	if(smaf && smaf->start)
	{
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(0, 128, 0)));

		MoveToEx(hdc, (smaf->start->time - smaf->start->duration) / PROption->RollWidth - scrx, 0, NULL);
		LineTo(hdc, (smaf->start->time - smaf->start->duration) / PROption->RollWidth - scrx, height);

		DeleteObject(SelectObject(hdc, hpold));
	}
	// End
	if(smaf && smaf->stop)
	{
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(128, 0, 128)));

		MoveToEx(hdc, smaf->stop->time / PROption->RollWidth - scrx, 0, NULL);
		LineTo(hdc, smaf->stop->time / PROption->RollWidth - scrx, height);

		DeleteObject(SelectObject(hdc, hpold));
	}

	// Current
	{
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(255, 0, 0)));

		MoveToEx(hdc, PROption->CurrentPos / PROption->RollWidth - scrx, 0, NULL);
		LineTo(hdc, PROption->CurrentPos / PROption->RollWidth - scrx, height);

		DeleteObject(SelectObject(hdc, hpold));
	}

	if(PROption->VisibleTrack != -1)
	{
		SIZE s;
		TCHAR ts[32];

		SaveDC(hdc);

		SetBkMode(hdc, OPAQUE);
		SetBkColor(hdc, RGB(255, 255, 255));
		SetTextColor(hdc, RGB(255, 0, 0));

		wsprintf(ts, "Viewing Track %d", PROption->VisibleTrack + 1);

		GetTextExtentPoint32(hdc, ts, lstrlen(ts), &s);
		TextOut(hdc, width - s.cx, height - s.cy, ts, lstrlen(ts));

		RestoreDC(hdc, -1);
	}
	if(PROption->ShowVisibleEventNum)
	{
		SIZE s;
		TCHAR ts[32];

		SaveDC(hdc);

		SetBkMode(hdc, OPAQUE);
		SetBkColor(hdc, RGB(255, 255, 255));
		SetTextColor(hdc, RGB(0, 0, 0));

		wsprintf(ts, "ProcessedEvents: %u Event(s)", ProcessEventNum);
		GetTextExtentPoint32(hdc, ts, lstrlen(ts), &s);
		TextOut(hdc, 0, height - s.cy * 3, ts, lstrlen(ts));

		wsprintf(ts, "ProcessedNoteEvents: %u Event(s)", ProcessNoteEventNum);
		TextOut(hdc, 0, height - s.cy * 2, ts, lstrlen(ts));

		wsprintf(ts, "VisibleEvents: %u Event(s)", VisibleEventNum);
		TextOut(hdc, 0, height - s.cy, ts, lstrlen(ts));

		RestoreDC(hdc, -1);
	}

	if(PROption->OptionString)
	{
		SaveDC(hdc);

		SetBkMode(hdc, OPAQUE);
		SetBkColor(hdc, RGB(255, 255, 255));
		SetTextColor(hdc, RGB(0, 0, 0));

		TextOut(hdc, 0, 0, PROption->OptionString, lstrlen(PROption->OptionString));

		RestoreDC(hdc, -1);
	}

	DeleteObject(SelectObject(hdc, hfold));

	return;
}

static void Resize(HWND hWnd)
{
	// If you want window-client size, use GetClientRect instead of lParam.

	SCROLLINFO si;

	si.cbSize = sizeof(si);
	si.fMask = SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL;
	si.nMin = 0;
	si.nMax = 65535;
	si.nPage = 10;
	SetScrollInfo(hWnd, SB_HORZ, &si, TRUE);

	si.nMin = 0;
	si.nMax = 128 * PROLLOPTION.RollHeight;
	si.nPage = PROLLOPTION.RollHeight;
	SetScrollInfo(hWnd, SB_VERT, &si, TRUE);

	InvalidateRect(hWnd, NULL, FALSE);

	return;
}

EVENT* EventHittest(HWND hWnd, UINT time, int note)
{
	EVENT* e;
	EVENT* lastselect = NULL;

	for(e = PROLLOPTION.smaf->events; e; e = e->next)
	{
		BYTE st = e->status & 0xF0;

		if(time < e->time)
			break;

		if(st == 0x80 || st == 0x90)
		{
			if(PROLLOPTION.VisibleTrack == -1 || (e->status & 0x0F) == PROLLOPTION.VisibleTrack)
			{
				DWORD gt = getGateTime(e);
				if(e->data[0] == note && e->time <= time && time < (e->time + gt))
					lastselect = e;
			}
		}
	}

	return lastselect;
}

void SetEOS(SMAF* smaf)
{
	EVENT* ev;
	EVENT* eos;

	for(eos = smaf->events; eos; eos = eos->next)
	{
		// EOS
		if(eos->status == 0xFF && eos->data[0] == 0x2F && eos->data[1] == 0x00)
			break;
	}
	if(!eos)
	{
		eos = halloc(sizeof(EVENT));
		eos->status = 0xFF;
		eos->size = 2;
		eos->data = halloc(eos->size);
		eos->data[0] = 0x2F;
		eos->data[1] = 0x00;
	}else
		CutoutEvent(smaf, eos);

	for(ev = smaf->events; ev && ev->next; ev = ev->next)
		;

	if(!ev)
	{
		eos->time = 0;
	}else
	{
		eos->time = ev->time;
		if(((ev->status & 0xF0) == 0x80) || ((ev->status & 0xF0) == 0x90))
			eos->time += GetVVal(ev->data + (((ev->status & 0xF0) == 0x80) ? 1 : 2));
	}

	smaf->start = InsertEvent(smaf, eos);

	return;
}

LRESULT CALLBACK PRollWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_NCCREATE:
		{
			DWORD wex;

			wex = GetWindowLong(hWnd, GWL_EXSTYLE);
			wex |= WS_EX_CLIENTEDGE;
			SetWindowLong(hWnd, GWL_EXSTYLE, wex);
		}
		break;
	case WM_CREATE:
		SetWindowLong(hWnd, 0, (LONG)halloc(sizeof(PROLLOPTIONSTRUCT)));

		SendMessage(hWnd, WM_PROLL, PROLLWM_RELOADCONFIG, 0);
		PROLLOPTION.VisibleTrack = -1;
		PROLLOPTION.CurrentPos = 0;
		PROLLOPTION.smaf = NULL;
		PROLLOPTION.OptionString = NULL;
		PROLLOPTION.SoundingNote = -1;
		PROLLOPTION.HilightedEvent = NULL;
		PROLLOPTION.MovingEvent = NULL;
		PROLLOPTION.SizingEvent = NULL;
		PROLLOPTION.ReadOnly = FALSE;
		PROLLOPTION.MouseMovingType = MMT_NONE;
#ifdef PROLL_ENABLESKIPFIRST
		PROLLOPTION.PrePosition = 0;
		PROLLOPTION.PreFirstEvent = NULL;
#endif
		PROLLOPTION.ShowVisibleEventNum = FALSE;

		break;
	case WM_DESTROY:
		hfree((void*)GetWindowLong(hWnd, 0));
		break;
	case WM_MOUSEWHEEL:
		if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
		{
			if((short)HIWORD(wParam) > 0)
				PostMessage(hWnd, WM_HSCROLL, SB_PAGELEFT, 0);
			else
				PostMessage(hWnd, WM_HSCROLL, SB_PAGERIGHT, 0);
		}else
		{
			if((short)HIWORD(wParam) > 0)
				PostMessage(hWnd, WM_VSCROLL, SB_PAGEUP, 0);
			else
				PostMessage(hWnd, WM_VSCROLL, SB_PAGEDOWN, 0);
		}
		return 0;
		// DefWindowProcMsg]ꂽƂ]Ɠ]Ŗ[v
		//break;
	case WM_HSCROLL:
	case WM_VSCROLL:
		{
			SCROLLINFO scr;

			scr.cbSize = sizeof(scr);
			scr.fMask = SIF_ALL;
			if(msg == WM_HSCROLL)
				GetScrollInfo(hWnd, SB_HORZ, &scr);
			if(msg == WM_VSCROLL)
				GetScrollInfo(hWnd, SB_VERT, &scr);

			switch(LOWORD(wParam))
			{
			//case SB_TOP:
			case SB_LEFT:
				scr.nPos = scr.nMin;
				break;
			//case SB_BOTTOM:
			case SB_RIGHT:
				scr.nPos = scr.nMax;
				break;
			//case SB_LINEUP:
			case SB_LINELEFT:
				if(scr.nMin < scr.nPos)
					scr.nPos--;
				break;
			//case SB_LINEDOWN:
			case SB_LINERIGHT:
				if(scr.nPos < scr.nMax - 1)
					scr.nPos++;
				break;
			//case SB_PAGEUP:
			case SB_PAGELEFT:
				scr.nPos -= scr.nPage;
				break;
			//case SB_PAGEDOWN:
			case SB_PAGERIGHT:
				scr.nPos += scr.nPage;
				break;
			case SB_THUMBPOSITION:
			case SB_THUMBTRACK:
				scr.nPos = HIWORD(wParam);
				break;
			}

			scr.fMask = SIF_POS;
			if(msg == WM_HSCROLL)
				SetScrollInfo(hWnd, SB_HORZ, &scr, TRUE);
			if(msg == WM_VSCROLL)
				SetScrollInfo(hWnd, SB_VERT, &scr, TRUE);

			InvalidateRect(hWnd, NULL, FALSE);
		}

		return 0;
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc;
			RECT rc;

			HDC hdcbg;
			HBITMAP oldbmp;
			BITMAPINFO bi = {0};

			hdc = BeginPaint(hWnd, &ps);
			GetClientRect(hWnd, &rc);

			switch(MainOption.DrawFunc)
			{
			case 0:
				break;
			case 1:
				hdcbg = hdc;
				hdc = CreateCompatibleDC(hdcbg);

				bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
				bi.bmiHeader.biPlanes = 1;
				bi.bmiHeader.biWidth = rc.right;
				bi.bmiHeader.biHeight = rc.bottom;
				bi.bmiHeader.biBitCount = 24;
				oldbmp = SelectObject(hdc, CreateDIBSection(hdc, &bi, DIB_RGB_COLORS, NULL, NULL, 0));

				break;
			case 2:
				hdcbg = hdc;
				hdc = CreateCompatibleDC(hdcbg);

				oldbmp = SelectObject(hdc, CreateCompatibleBitmap(hdcbg, rc.right, rc.bottom));

				break;
			}

			draw(hWnd, hdc, rc.right, rc.bottom);

			switch(MainOption.DrawFunc)
			{
			case 0:
				break;
			case 1:
			case 2:
				BitBlt(hdcbg, 0, 0, rc.right, rc.bottom, hdc, 0, 0, SRCCOPY);

				DeleteObject(SelectObject(hdc, oldbmp));
				DeleteDC(hdc);

				break;
			}

			EndPaint(hWnd, &ps);
		}
		break;
	case WM_SIZE:
		Resize(hWnd);
		break;
	case WM_PROLL:
		switch(LOWORD(wParam))
		{
		//=== SYSTEM ===//
		case PROLLWM_RELOADCONFIG:
			PROLLOPTION.RollWidth = MainOption.PRollOption.RollWidth;
			PROLLOPTION.RollHeight = MainOption.PRollOption.RollHeight;
			PROLLOPTION.NiagaraTouch = MainOption.PRollOption.NiagaraTouch;
			break;

		//=== OPTION/COMMAND ===//
		case PROLLWM_SCROLLTO:
			SetScrollPos(hWnd, SB_HORZ, (int)lParam / (int)PROLLOPTION.RollWidth + (short)HIWORD(wParam), TRUE);
			break;
		case PROLLWM_SETROLLWIDTH:
			{
				UINT time = SendMessage(hWnd, WM_PROLL, PROLLWM_GETSCROLL, 0);
				PROLLOPTION.RollWidth = lParam;
				SendMessage(hWnd, WM_PROLL, PROLLWM_SCROLLTO, time);
			}
			break;
		case PROLLWM_SETROLLHEIGHT:
			PROLLOPTION.RollHeight = lParam;
			Resize(hWnd);
			break;
		case PROLLWM_SETVISIBLETRACK:
			PROLLOPTION.VisibleTrack = lParam;
			break;
		case PROLLWM_GETROLLWIDTH:
			return PROLLOPTION.RollWidth;
		case PROLLWM_GETROLLHEIGHT:
			return PROLLOPTION.RollHeight;
		case PROLLWM_GETVISIBLETRACK:
			return PROLLOPTION.VisibleTrack;
		case PROLLWM_GETSCROLL:
			return GetScrollPos(hWnd, SB_HORZ) * PROLLOPTION.RollWidth;
		case PROLLWM_SETSMAF:
			PROLLOPTION.HilightedEvent = NULL;
			PROLLOPTION.smaf = (SMAF*)lParam;
			InvalidateRect(hWnd, NULL, FALSE);
			break;
		case PROLLWM_SETCURRENTPOS:
			PROLLOPTION.CurrentPos = lParam;
			InvalidateRect(hWnd, NULL, FALSE);
			break;
		case PROLLWM_SETOPTIONSTRING:
			PROLLOPTION.OptionString = (LPCTSTR)lParam;
			break;
		case PROLLWM_GETOPTIONSTRING:
			return (LRESULT)PROLLOPTION.OptionString;
		case PROLLWM_SETREADONLY:
			PROLLOPTION.ReadOnly = (BOOL)lParam;
			break;
		case PROLLWM_SETHILIGHTEVENT:
			PROLLOPTION.HilightedEvent = (EVENT*)lParam;
			break;
		case PROLLWM_SETNIAGARATOUCH:
			PROLLOPTION.NiagaraTouch = (BOOL)lParam;
			break;
		case PROLLWM_GETNIAGARATOUCH:
			return PROLLOPTION.NiagaraTouch;
		case PROLLWM_GETREADONLY:
			return PROLLOPTION.ReadOnly;
		case PROLLWM_SETSHOWVISIBLEEVENTNUM:
			PROLLOPTION.ShowVisibleEventNum = (BOOL)lParam;
			break;
		case PROLLWM_GETSHOWVISIBLEEVENTNUM:
			return PROLLOPTION.ShowVisibleEventNum;
		}
		break;
	case WM_LBUTTONDOWN:
		SetCapture(hWnd);
		if(HIWORD(lParam) < POSBARHEIGHT)
		{
			if(PROLLOPTION.smaf)
			{
				PROLLOPTION.CurrentPos = (GetScrollPos(hWnd, SB_HORZ) + LOWORD(lParam)) * PROLLOPTION.RollWidth;
				SendMessage(GetParent(hWnd), WM_PROLLNOTIFY, PN_CURRENTPOSCHANGED, PROLLOPTION.CurrentPos);
				InvalidateRect(hWnd, NULL, FALSE);
				PROLLOPTION.MouseMovingType = MMT_POSBAR;
			}
		}
		if(POSBARHEIGHT <= HIWORD(lParam))
		{
			UINT x = (GetScrollPos(hWnd, SB_HORZ) + LOWORD(lParam)) * PROLLOPTION.RollWidth;
			int y = 127 - ((HIWORD(lParam) + GetScrollPos(hWnd, SB_VERT) - POSBARHEIGHT) / PROLLOPTION.RollHeight);

			if(0 <= y && y <= 127)
			{
				if(PROLLOPTION.smaf)
				{
					PROLLOPTION.HilightedEvent = EventHittest(hWnd, x, y);
					if(!PROLLOPTION.ReadOnly && PROLLOPTION.HilightedEvent)
					{
						if(
							(PROLLOPTION.HilightedEvent->time + (PROLLOPTION.HilightedEvent->time + getGateTime(PROLLOPTION.HilightedEvent)) * 4) / 5 < x
							|| (PROLLOPTION.HilightedEvent->time + getGateTime(PROLLOPTION.HilightedEvent) - 2) < x
							)
						{
							POINT p;

							PROLLOPTION.SizingEvent = PROLLOPTION.HilightedEvent;

							p.x = (PROLLOPTION.SizingEvent->time + getGateTime(PROLLOPTION.SizingEvent)) / PROLLOPTION.RollWidth - GetScrollPos(hWnd, SB_HORZ);
							p.y = HIWORD(lParam);
							ClientToScreen(hWnd, &p);
							SetCursorPos(p.x, p.y);
						}else
						{
							PROLLOPTION.MovingEvent = PROLLOPTION.HilightedEvent;
							PROLLOPTION.MovingPosDelta = PROLLOPTION.MovingEvent->time - x;
						}
					}
					InvalidateRect(hWnd, NULL, FALSE);
				}

				PROLLOPTION.SoundingNote = y;
				SendMessage(GetParent(hWnd), WM_PROLLNOTIFY, MAKEWPARAM(PN_CLICKED, y), x);

				PROLLOPTION.MouseMovingType = MMT_ONROLL;
			}
		}
		break;
	case WM_LBUTTONDBLCLK:
		if(HIWORD(lParam) >= POSBARHEIGHT)
		{
			UINT x = (GetScrollPos(hWnd, SB_HORZ) + LOWORD(lParam)) * PROLLOPTION.RollWidth;
			int y = 127 - ((HIWORD(lParam) + GetScrollPos(hWnd, SB_VERT) - POSBARHEIGHT) / PROLLOPTION.RollHeight);

			if(PROLLOPTION.smaf && 0 <= y && y <= 127)
			{
				if(EventHittest(hWnd, x, y))
				{
					// setting, menu, etc.
					;
				}else
				{
					if(!PROLLOPTION.ReadOnly)
					{
						// blank space->create note
						EVENT* e;
						DWORD gt = 500;
						UINT tr = (PROLLOPTION.VisibleTrack == -1) ? 0 : PROLLOPTION.VisibleTrack;

						e = halloc(sizeof(EVENT));
						e->time = x / PROLLOPTION.RollWidth * PROLLOPTION.RollWidth;	// aligned placing
						e->status = 0x90 | tr;
						e->size = 2 + SetVValSize(gt);
						e->data = halloc(e->size);
						e->data[0] = y;
						e->data[1] = 0x7F;	// velo
						SetVVal(e->data + 2, gt);

						PROLLOPTION.smaf->start = InsertEvent(PROLLOPTION.smaf, e);
						SetEOS(PROLLOPTION.smaf);

						InvalidateRect(hWnd, NULL, FALSE);
					}
				}
			}

			if(0 <= y && y <= 127)
				SendMessage(GetParent(hWnd), WM_PROLLNOTIFY, MAKEWPARAM(PN_DBLCLICKED, y), (GetScrollPos(hWnd, SB_HORZ) + LOWORD(lParam)) * PROLLOPTION.RollWidth);
		}
		break;
	case WM_MOUSEMOVE:
		if(GetCapture() == hWnd)	// pressing...tenuki
		{
			if(PROLLOPTION.MouseMovingType == MMT_POSBAR)
			{
				int x = (int)(GetScrollPos(hWnd, SB_HORZ) + (short)LOWORD(lParam)) * (int)PROLLOPTION.RollWidth;

				if(x < 0)
					x = 0;

				PROLLOPTION.CurrentPos = x;
				SendMessage(GetParent(hWnd), WM_PROLLNOTIFY, PN_CURRENTPOSCHANGED, PROLLOPTION.CurrentPos);
				InvalidateRect(hWnd, NULL, FALSE);
			}
			if(PROLLOPTION.MouseMovingType == MMT_ONROLL && POSBARHEIGHT <= HIWORD(lParam))
			{
				int x = (int)(GetScrollPos(hWnd, SB_HORZ) + (int)(short)LOWORD(lParam)) * (int)PROLLOPTION.RollWidth;
				int y = 127 - ((HIWORD(lParam) + GetScrollPos(hWnd, SB_VERT) - POSBARHEIGHT) / PROLLOPTION.RollHeight);

				if(x < 0)
					x = 0;

				if(PROLLOPTION.MovingEvent)
				{
					// align: for "duration-base dividing in convert" problem
					int time = (x + PROLLOPTION.MovingPosDelta) / PROLLOPTION.smaf->dbase * PROLLOPTION.smaf->dbase;

					if(time < 0)
						time = 0;

					PROLLOPTION.MovingEvent->time = time;
					// refresh duration
					CutoutEvent(PROLLOPTION.smaf, PROLLOPTION.MovingEvent);
					PROLLOPTION.smaf->start = InsertEvent(PROLLOPTION.smaf, PROLLOPTION.MovingEvent);

					if(0 <= y && y <= 127)
						PROLLOPTION.MovingEvent->data[0] = y;
					SetEOS(PROLLOPTION.smaf);

					InvalidateRect(hWnd, NULL, FALSE);
				}
				if(PROLLOPTION.SizingEvent)
				{
					EVENT* e = PROLLOPTION.SizingEvent;

					if((e->status & 0xF0) == 0x80)
					{
						BYTE note = e->data[0];
						int newgt = x - e->time;

						if(newgt < (int)PROLLOPTION.smaf->gbase)
							newgt = PROLLOPTION.smaf->gbase;

						// align
						newgt = newgt / PROLLOPTION.smaf->gbase * PROLLOPTION.smaf->gbase;

						hfree(e->data);
						e->size = 1 + SetVValSize(newgt);
						e->data = halloc(e->size);
						e->data[0] = note;
						SetVVal(e->data + 1, newgt);
					}
					if((e->status & 0xF0) == 0x90)
					{
						BYTE note = e->data[0];
						BYTE vel = e->data[1];
						int newgt = x - e->time;

						if(newgt < (int)PROLLOPTION.smaf->gbase)
							newgt = PROLLOPTION.smaf->gbase;

						// align
						newgt = newgt / PROLLOPTION.smaf->gbase * PROLLOPTION.smaf->gbase;

						hfree(e->data);
						e->size = 2 + SetVValSize(newgt);
						e->data = halloc(e->size);
						e->data[0] = note;
						e->data[1] = vel;
						SetVVal(e->data + 2, newgt);
					}

					SetEOS(PROLLOPTION.smaf);

					InvalidateRect(hWnd, NULL, FALSE);
				}
				if(PROLLOPTION.NiagaraTouch && 0 <= y && y <= 127 && PROLLOPTION.SoundingNote != y)
				{
					SendMessage(GetParent(hWnd), WM_PROLLNOTIFY, PN_RELEASED, 0);
					PROLLOPTION.SoundingNote = y;
					SendMessage(GetParent(hWnd), WM_PROLLNOTIFY, MAKEWPARAM(PN_CLICKED, y), x);
				}
			}
		}
		break;
	case WM_LBUTTONUP:
		ReleaseCapture();
		if(PROLLOPTION.MovingEvent || PROLLOPTION.SizingEvent)
		{
			PROLLOPTION.MovingEvent = NULL;
			PROLLOPTION.SizingEvent = NULL;
			InvalidateRect(hWnd, NULL, FALSE);
		}
		SendMessage(GetParent(hWnd), WM_PROLLNOTIFY, PN_RELEASED, 0);
		PROLLOPTION.SoundingNote = -1;
		PROLLOPTION.MouseMovingType = MMT_NONE;
		break;
	case WM_MBUTTONDOWN:
		if(HIWORD(lParam) >= POSBARHEIGHT)
		{
			UINT x = (GetScrollPos(hWnd, SB_HORZ) + LOWORD(lParam)) * PROLLOPTION.RollWidth;
			int y = 127 - ((HIWORD(lParam) + GetScrollPos(hWnd, SB_VERT) - POSBARHEIGHT) / PROLLOPTION.RollHeight);

			if(0 <= y && y <= 127)
				SendMessage(GetParent(hWnd), WM_PROLLNOTIFY, MAKEWPARAM(PN_WHEELCLICKED, y), x);
		}
		return 0;
	case WM_RBUTTONDOWN:
		//폜
		if(HIWORD(lParam) >= POSBARHEIGHT)
		{
			UINT x = (GetScrollPos(hWnd, SB_HORZ) + LOWORD(lParam)) * PROLLOPTION.RollWidth;
			int y = 127 - ((HIWORD(lParam) + GetScrollPos(hWnd, SB_VERT) - POSBARHEIGHT) / PROLLOPTION.RollHeight);

			if(0 <= y && y <= 127)
			{
				if(PROLLOPTION.smaf)
				{
					EVENT* e;
					if(e = EventHittest(hWnd, x, y))
					{
						if(!PROLLOPTION.ReadOnly)
						{
							CutoutEvent(PROLLOPTION.smaf, e);
							hfree(e->data);
							hfree(e);
							PROLLOPTION.HilightedEvent = NULL;
						}else
							PROLLOPTION.HilightedEvent = e;

						InvalidateRect(hWnd, NULL, FALSE);
					}
				}
				//SendMessage(GetParent(hWnd), WM_PROLLNOTIFY, MAKEWPARAM(PN_CONTEXTMENU, y), x);
			}
		}
		break;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

BOOL PRollInitialize(void)
{
	WNDCLASSEX wc = {0};

	wc.cbSize = sizeof(wc);
	wc.cbWndExtra = sizeof(PROLLOPTIONSTRUCT*);
	// Ƃ...RedrawBkgndœhԂȂȂ邩B
	//wc.hbrBackground = (HBRUSH)(COLOR_3DFACE + 1);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hInstance = GetModuleHandle(NULL);
	wc.hIcon = LoadIcon(wc.hInstance, MAKEINTRESOURCE(IDI_MAIN));
	wc.lpfnWndProc = PRollWndProc;
	wc.lpszClassName = "MMFToolPRollWC";
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MAIN);
	wc.style = CS_DBLCLKS;

	if(!RegisterClassEx(&wc))
		return FALSE;

	return TRUE;
}
